---------------------------------------------------------------------------
--
-- Copyright (c) 2016
-- Mitsubishi Electric Corporation
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are
-- met:
-- 
-- 1. Redistributions of source code must retain the above copyright
-- notice, this list of conditions and the following disclaimer.
-- 
-- 2. Redistributions in binary form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-- "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-- LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-- A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-- HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-- SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-- LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-- THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-- (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-- OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--
-- File Name: CipherCore.vhd
-- Type: RTL 
-- Author: Takeshi SUGAWARA
-- Date: 2016/06/21
-- Version: 1.0
-- Abstract:
-- Modification History: 
--
---------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
  component MA_minalpher is
    port(
      Din:  in  std_logic_vector(255 downto 0);
      Kin:  in  std_logic_vector(127 downto 0);
      Nin:  in  std_logic_vector(127 downto 0);  
      Dout: out std_logic_vector(255 downto 0);
      Tout: out std_logic_vector(255 downto 0);
      command:   in std_logic_vector(4 downto 0);
      CLK:       in std_logic;
      RSTn:      in std_logic;
      activate:  in std_logic;
      Dout_valid: out std_logic;
      Tout_valid: out std_logic;
      command_finish: out std_logic
      );
  end component;

type mystate_type is (ST_IDLE,
                      ST_NPUB1,
                      ST_NPUB2,
                      ST_NPUB3,
                      ST_AD_INIT1,
                      ST_AD_INIT2, 
                      ST_AD1,
                      ST_AD2,
                      ST_AD3,
                      ST_AD4,
                      ST_AD_FIN1,
                      ST_AD_FIN2,
                      ST_AD_FIN3,
                      ST_AD_FIN4, 
                      ST_ENCDEC_INIT1,
                      ST_ENCDEC_INIT2,
                      ST_ENCDEC_INIT3,
                      ST_ENCDEC_INIT4, 
                      ST_ENC1,
                      ST_ENC2,
                      ST_ENC3,
                      ST_DEC1,
                      ST_DEC2,
                      ST_DEC3,                         
                      ST_ENC_FIN1,
                      ST_ENC_FIN2, 
                      ST_DEC_FIN1,
                      ST_DEC_FIN2,
                      ST_DEC_COMPARE,
                      ST_KEY);
  signal state_reg: mystate_type;

  signal ma_Din:            std_logic_vector(255 downto 0);
  signal ma_Kin:            std_logic_vector(127 downto 0);
  signal ma_Nin:            std_logic_vector(127 downto 0);  
  signal ma_Dout:           std_logic_vector(255 downto 0);
  signal ma_Tout:           std_logic_vector(255 downto 0);
  signal calculated_tag:    std_logic_vector(127 downto 0);
  signal ma_command:        std_logic_vector(4 downto 0);
  signal ma_CLK:            std_logic;
  signal ma_RSTn:           std_logic;
  signal ma_activate:       std_logic;
  signal ma_Dout_valid:     std_logic;
  signal ma_Tout_valid:     std_logic;
  signal ma_command_finish: std_logic;
  signal is_padded:         std_logic;
  
  constant COM_KEYSET             : std_logic_vector(4 downto 0) := "00000";
  constant COM_NSET               : std_logic_vector(4 downto 0) := "00001";
  constant COM_TWEAK_ENCMAC       : std_logic_vector(4 downto 0) := "00010";
  constant COM_TWEAK_ENCENC       : std_logic_vector(4 downto 0) := "00011";
  constant COM_TWEAK_2            : std_logic_vector(4 downto 0) := "00101";
  constant COM_TWEAK_3            : std_logic_vector(4 downto 0) := "00110";
  constant COM_RESET_TAG          : std_logic_vector(4 downto 0) := "00111";
  constant COM_TAG                : std_logic_vector(4 downto 0) := "01000";
  constant COM_TAG_FINAL          : std_logic_vector(4 downto 0) := "01001";
  constant COM_ENC_THEN_TAG       : std_logic_vector(4 downto 0) := "01101";
  constant COM_ENC_THEN_TAG_FINAL : std_logic_vector(4 downto 0) := "01110";
  constant COM_DEC_THEN_TAG       : std_logic_vector(4 downto 0) := "10001";
  constant COM_DEC_THEN_TAG_FINAL : std_logic_vector(4 downto 0) := "10010";
  constant COM_NOP                : std_logic_vector(4 downto 0) := "11111";

  signal expected_tag:       std_logic_vector(127 downto 0);
  signal bdo_tmp:            std_logic_vector(255 downto 0);
  signal bdo_size_tmp:       std_logic_vector(  5 downto 0);
  signal bdo_compress:       std_logic_vector( 31 downto 0);
  signal bdo_compress_reg:   std_logic_vector( 31 downto 0);
  signal ma_Tout_valid_mask: std_logic;
  signal bdi_is_npub:        std_logic;
  signal bdi_is_ad:          std_logic;  
begin
  minalpher: MA_minalpher port map(
    Din => ma_Din,
    Kin => ma_Kin,
    Nin => ma_Nin,
    Dout => ma_Dout,
    Tout => ma_Tout,
    CLK => ma_CLK,
    RSTn => ma_RSTn,
    command => ma_command,
    activate => ma_activate,
    Dout_valid => ma_Dout_valid,
    Tout_valid => ma_Tout_valid,
    command_finish => ma_command_finish);

  ma_CLK <= CLK;
  ma_RSTn <= not rst;
  ma_Nin <= bdi(255 downto 128);
  ma_Kin <= key;
  ma_Din <= bdi;

  calculated_tag <= ma_Tout(255 downto 128);  
  expected_tag <= bdi(255 downto 128);
  
  bdo_size <= "100000" when (state_reg /= ST_DEC_FIN2) else bdo_size_tmp;
  bdo_tmp <= ma_Dout;
  bdo <= bdo_tmp;
  key_ready <= '1' when (state_reg = ST_KEY) else '0';
  bdi_ready <= '1' when ( (state_reg = ST_ENC_FIN1   ) or
                          (state_reg = ST_ENC1       ) or
                          (state_reg = ST_DEC1       ) or
                          (state_reg = ST_AD_FIN3    ) or
                          (state_reg = ST_AD2        ) or
                          (state_reg = ST_DEC_FIN1   ) or    
                          (state_reg = ST_NPUB1      ) or       
                          (state_reg = ST_DEC_COMPARE) ) else
               '0';

  ma_Tout_valid_mask <= '0' when ((state_reg = ST_AD_FIN4) or (state_reg = ST_DEC_FIN2)) else
                        ma_Tout_valid;
  bdo_valid <= ma_Dout_valid or ma_Tout_valid_mask;
  
  msg_auth_valid <= '1' when ( (expected_tag = calculated_tag) and (state_reg = ST_DEC_COMPARE) ) else
                    '0';
  msg_auth_done <= '1' when (state_reg = ST_DEC_COMPARE) else '0';

  is_padded <= '0' when (bdi_pad_loc = X"00000000") else '1';
  bdi_is_npub <= '1' when (bdi_type = "110") else '0';
  bdi_is_ad   <= '1' when (bdi_type(2 downto 1) = "00") else '0';  
  
  ma_activate <= '1' when ( (state_reg = ST_KEY         ) or
                            (state_reg = ST_NPUB1       ) or
                            (state_reg = ST_NPUB2       ) or
                            (state_reg = ST_ENCDEC_INIT1) or
                            (state_reg = ST_ENCDEC_INIT3) or
                            (state_reg = ST_ENC1        ) or
                            (state_reg = ST_DEC1        ) or
                            (state_reg = ST_ENC_FIN1    ) or
                            (state_reg = ST_DEC_FIN1    ) or
                            (state_reg = ST_AD_INIT1    ) or
                            (state_reg = ST_AD_FIN1     ) or
                            (state_reg = ST_AD_FIN2     ) or
                            (state_reg = ST_AD_FIN3     ) or
                            (state_reg = ST_AD1         ) or
                            (state_reg = ST_AD2         ) ) else
                 '0';
               
  bdo_compress( 0) <= '1' when bdo_tmp(  7 downto   0) /= X"00" else '0';
  bdo_compress( 1) <= '1' when bdo_tmp( 15 downto   8) /= X"00" else '0';
  bdo_compress( 2) <= '1' when bdo_tmp( 23 downto  16) /= X"00" else '0';
  bdo_compress( 3) <= '1' when bdo_tmp( 31 downto  24) /= X"00" else '0';
  bdo_compress( 4) <= '1' when bdo_tmp( 39 downto  32) /= X"00" else '0';
  bdo_compress( 5) <= '1' when bdo_tmp( 47 downto  40) /= X"00" else '0';
  bdo_compress( 6) <= '1' when bdo_tmp( 55 downto  48) /= X"00" else '0';
  bdo_compress( 7) <= '1' when bdo_tmp( 63 downto  56) /= X"00" else '0';
  bdo_compress( 8) <= '1' when bdo_tmp( 71 downto  64) /= X"00" else '0';
  bdo_compress( 9) <= '1' when bdo_tmp( 79 downto  72) /= X"00" else '0';
  bdo_compress(10) <= '1' when bdo_tmp( 87 downto  80) /= X"00" else '0';
  bdo_compress(11) <= '1' when bdo_tmp( 95 downto  88) /= X"00" else '0';
  bdo_compress(12) <= '1' when bdo_tmp(103 downto  96) /= X"00" else '0';
  bdo_compress(13) <= '1' when bdo_tmp(111 downto 104) /= X"00" else '0';
  bdo_compress(14) <= '1' when bdo_tmp(119 downto 112) /= X"00" else '0';
  bdo_compress(15) <= '1' when bdo_tmp(127 downto 120) /= X"00" else '0';
  bdo_compress(16) <= '1' when bdo_tmp(135 downto 128) /= X"00" else '0';
  bdo_compress(17) <= '1' when bdo_tmp(143 downto 136) /= X"00" else '0';
  bdo_compress(18) <= '1' when bdo_tmp(151 downto 144) /= X"00" else '0';
  bdo_compress(19) <= '1' when bdo_tmp(159 downto 152) /= X"00" else '0';
  bdo_compress(20) <= '1' when bdo_tmp(167 downto 160) /= X"00" else '0';
  bdo_compress(21) <= '1' when bdo_tmp(175 downto 168) /= X"00" else '0';
  bdo_compress(22) <= '1' when bdo_tmp(183 downto 176) /= X"00" else '0';
  bdo_compress(23) <= '1' when bdo_tmp(191 downto 184) /= X"00" else '0';
  bdo_compress(24) <= '1' when bdo_tmp(199 downto 192) /= X"00" else '0';
  bdo_compress(25) <= '1' when bdo_tmp(207 downto 200) /= X"00" else '0';
  bdo_compress(26) <= '1' when bdo_tmp(215 downto 208) /= X"00" else '0';
  bdo_compress(27) <= '1' when bdo_tmp(223 downto 216) /= X"00" else '0';
  bdo_compress(28) <= '1' when bdo_tmp(231 downto 224) /= X"00" else '0';
  bdo_compress(29) <= '1' when bdo_tmp(239 downto 232) /= X"00" else '0';
  bdo_compress(30) <= '1' when bdo_tmp(247 downto 240) /= X"00" else '0';
  bdo_compress(31) <= '1' when bdo_tmp(255 downto 248) /= X"00" else '0';  
  
  proc_priority_encoder: process(bdo_compress_reg)
  begin
    if    bdo_compress_reg( 0) = '1' then bdo_size_tmp <= "011111";
    elsif bdo_compress_reg( 1) = '1' then bdo_size_tmp <= "011110";
    elsif bdo_compress_reg( 2) = '1' then bdo_size_tmp <= "011101";
    elsif bdo_compress_reg( 3) = '1' then bdo_size_tmp <= "011100";
    elsif bdo_compress_reg( 4) = '1' then bdo_size_tmp <= "011011";
    elsif bdo_compress_reg( 5) = '1' then bdo_size_tmp <= "011010";
    elsif bdo_compress_reg( 6) = '1' then bdo_size_tmp <= "011001";
    elsif bdo_compress_reg( 7) = '1' then bdo_size_tmp <= "011000";
    elsif bdo_compress_reg( 8) = '1' then bdo_size_tmp <= "010111";
    elsif bdo_compress_reg( 9) = '1' then bdo_size_tmp <= "010110";
    elsif bdo_compress_reg(10) = '1' then bdo_size_tmp <= "010101";
    elsif bdo_compress_reg(11) = '1' then bdo_size_tmp <= "010100";
    elsif bdo_compress_reg(12) = '1' then bdo_size_tmp <= "010011";
    elsif bdo_compress_reg(13) = '1' then bdo_size_tmp <= "010010";
    elsif bdo_compress_reg(14) = '1' then bdo_size_tmp <= "010001";
    elsif bdo_compress_reg(15) = '1' then bdo_size_tmp <= "010000";
    elsif bdo_compress_reg(16) = '1' then bdo_size_tmp <= "001111";
    elsif bdo_compress_reg(17) = '1' then bdo_size_tmp <= "001110";
    elsif bdo_compress_reg(18) = '1' then bdo_size_tmp <= "001101";
    elsif bdo_compress_reg(19) = '1' then bdo_size_tmp <= "001100";
    elsif bdo_compress_reg(20) = '1' then bdo_size_tmp <= "001011";
    elsif bdo_compress_reg(21) = '1' then bdo_size_tmp <= "001010";
    elsif bdo_compress_reg(22) = '1' then bdo_size_tmp <= "001001";
    elsif bdo_compress_reg(23) = '1' then bdo_size_tmp <= "001000";
    elsif bdo_compress_reg(24) = '1' then bdo_size_tmp <= "000111";
    elsif bdo_compress_reg(25) = '1' then bdo_size_tmp <= "000110";
    elsif bdo_compress_reg(26) = '1' then bdo_size_tmp <= "000101";
    elsif bdo_compress_reg(27) = '1' then bdo_size_tmp <= "000100";
    elsif bdo_compress_reg(28) = '1' then bdo_size_tmp <= "000011";
    elsif bdo_compress_reg(29) = '1' then bdo_size_tmp <= "000010";
    elsif bdo_compress_reg(30) = '1' then bdo_size_tmp <= "000001";
    else                                  bdo_size_tmp <= "000000";
    end if;
  end process proc_priority_encoder;
    
  proc_command_decoder: process(state_reg)
  begin
    case state_reg is
      when ST_KEY          => ma_command <= COM_KEYSET;
      when ST_NPUB1        => ma_command <= COM_NSET;
      when ST_NPUB2        => ma_command <= COM_RESET_TAG;
      when ST_ENCDEC_INIT1 => ma_command <= COM_TWEAK_ENCENC;
      when ST_ENCDEC_INIT3 => ma_command <= COM_TWEAK_2;
      when ST_AD_INIT1     => ma_command <= COM_TWEAK_ENCMAC;
      when ST_AD_FIN1      => ma_command <= COM_TWEAK_3;
      when ST_AD_FIN2      => ma_command <= COM_TWEAK_3;
      when ST_AD_FIN3      => ma_command <= COM_TAG_FINAL;
      when ST_ENC1         => ma_command <= COM_ENC_THEN_TAG;            
      when ST_DEC1         => ma_command <= COM_DEC_THEN_TAG;
      when ST_ENC_FIN1     => ma_command <= COM_ENC_THEN_TAG_FINAL;
      when ST_DEC_FIN1     => ma_command <= COM_DEC_THEN_TAG_FINAL;
      when ST_AD1          => ma_command <= COM_TWEAK_2;
      when ST_AD2          => ma_command <= COM_TAG;
      when others          => ma_command <= COM_NOP;
    end case;
  end process proc_command_decoder;

  proc_bdo_compress_reg: process(clk)
  begin
    if rising_edge( clk ) then
      bdo_compress_reg <= bdo_compress;
    end if;
  end process proc_bdo_compress_reg;
  
  process(clk)
  begin
    if rising_edge( clk ) then
      if rst = '1' then
        state_reg <= ST_IDLE;
      else
        case state_reg is
          when ST_IDLE =>          
            if ((key_valid = '1') and (key_update = '1')) then
              state_reg <= ST_KEY;
            elsif ((bdi_valid = '1') and (bdi_is_npub = '1')) then
              state_reg <= ST_NPUB1; end if;

          when ST_KEY   => state_reg <= ST_IDLE;
          when ST_NPUB1 => state_reg <= ST_NPUB2;
          when ST_NPUB2 => state_reg <= ST_NPUB3;
          when ST_NPUB3 => -- wait for next operation
            if(bdi_valid = '1') then              
              if bdi_is_ad = '1' then              
                state_reg <= ST_AD_INIT1;
              else
                state_reg <= ST_ENCDEC_INIT1;
              end if;
            end if;

          when ST_AD_INIT1 => state_reg <= ST_AD_INIT2;
          when ST_AD_INIT2 =>
            if(ma_command_finish = '1') then
              if(bdi_eot = '1') then
                state_reg <= ST_AD_FIN1;
              else
                state_reg <= ST_AD1;
              end if;
            end if;
              
          when ST_AD1 => state_reg <= ST_AD2; -- x2
          when ST_AD2 => state_reg <= ST_AD3; -- tag   
          when ST_AD3 => 
            if(ma_command_finish = '1') then
              if(bdi_valid = '1') then              
                if(bdi_eot = '1') then state_reg <= ST_AD_FIN1;
                else                   state_reg <= ST_AD1; end if;
              else
                state_reg <= ST_AD4;
              end if;
            end if;
          when ST_AD4 =>
            if(bdi_valid = '1') then              
              if(bdi_eot = '1') then state_reg <= ST_AD_FIN1;
              else                   state_reg <= ST_AD1; end if;
            end if;

          when ST_AD_FIN1 =>
            if(is_padded = '1') then state_reg <= ST_AD_FIN2;
            else                     state_reg <= ST_AD_FIN3; end if;
          when ST_AD_FIN2 => state_reg <= ST_AD_FIN3;
          when ST_AD_FIN3 => state_reg <= ST_AD_FIN4;
          when ST_AD_FIN4 =>                             
            if(ma_command_finish = '1') then
              state_reg <= ST_ENCDEC_INIT1;
            end if;
            
          when ST_ENCDEC_INIT1 => state_reg <= ST_ENCDEC_INIT2;
          when ST_ENCDEC_INIT2 =>
            if(ma_command_finish = '1') then
              state_reg <= ST_ENCDEC_INIT3;
            end if;
          when ST_ENCDEC_INIT3 =>
            if(bdi_valid = '1') then              
              if    (decrypt = '0' and bdi_eot = '1') then state_reg <= ST_ENC_FIN1;
              elsif (decrypt = '0' and bdi_eot = '0') then state_reg <= ST_ENC1;
              elsif (decrypt = '1' and bdi_eot = '1') then state_reg <= ST_DEC_FIN1;
              else                                         state_reg <= ST_DEC1;
              end if;
            end if;

          when ST_ENC1 => state_reg <= ST_ENC2;
          when ST_ENC2 =>
            if(ma_command_finish = '1') then
              if(bdi_valid = '1') then              
                if(bdi_eot = '1') then state_reg <= ST_ENC_FIN1;
                else                   state_reg <= ST_ENC1; end if;
              else
                state_reg <= ST_ENC3;
              end if;
            end if;
          when ST_ENC3 =>
            if(bdi_valid = '1') then              
              if(bdi_eot = '1') then state_reg <= ST_ENC_FIN1;              
              else                   state_reg <= ST_ENC1; end if;
            end if;

          when ST_DEC1 => state_reg <= ST_DEC2;
          when ST_DEC2 =>
            if(ma_command_finish = '1') then
              if(bdi_valid = '1') then              
                if(bdi_eot = '1') then state_reg <= ST_DEC_FIN1;
                else                   state_reg <= ST_DEC1; end if;
              else
                state_reg <= ST_DEC3;
              end if;
            end if;
          when ST_DEC3 =>
            if(bdi_valid = '1') then              
              if(bdi_eot = '1') then state_reg <= ST_DEC_FIN1;
              else                   state_reg <= ST_DEC1; end if;
            end if;
          when ST_ENC_FIN1 => state_reg <= ST_ENC_FIN2;
          when ST_ENC_FIN2 =>
            if(ma_command_finish = '1') then
              state_reg <= ST_IDLE;
            end if;
          
          when ST_DEC_FIN1 => state_reg <= ST_DEC_FIN2;
          when ST_DEC_FIN2 =>
            if(ma_command_finish = '1') then
              state_reg <= ST_DEC_COMPARE;
            end if;
          when ST_DEC_COMPARE => state_reg <= ST_IDLE;
          when others => -- do nothing
        end case;
      end if;
    end if;
  end process;        
end structure;
